#ifdef HAVE_CONFIG_H
#include <config.hpp>
#endif

#include "mc-hid-server.hpp"
#include "log/log.hpp"

#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <stdio.h>
#include <limits.h>
#include <pthread.h>


// For getting current time
#include <time.h>
#ifndef CLOCK_MONOTIC
#   define CLOCK_MONOTIC CLOCK_REALTIME
#endif

#define NOISE_TIMEOUT_MS    400
#define GPIO_INT_PIN        4
#define MCP23017_ADR        0x20

#define BTN_ESCAPE          0x0001
#define BTN_FAVORITES       0x0002
#define BTN_RECENT          0x0004

#define BTN_REWIND          0x0008
#define BTN_PLAYPAUSE       0x0010
#define BTN_FASTFORWARD     0x0020

#define BTN_CANCEL          0x0040
#define BTN_OK              0x0080
#define BTN_UP              0x0100
#define BTN_DOWN            0x0200

#define SENSE_MINIJACK      0x0400

#define PIN_RED             12
#define PIN_GREEN           13
#define PIN_BLUE            14

#define PULSE_STEPS         32
#define PULSE_WAIT_US       20000

#define TRUE                1
#define FALSE               0

using namespace std;

static const char *HID_SERVER_NAME = "nl.miqra.MediaCore.Hid";
static const char *HID_SERVER_PATH = "/nl/miqra/MediaCore/Hid";


// function to get current time in ms
int64_t now_ms(void);
// signal handler
void niam(int sig);

HidServer::HidServer(DBus::Connection &connection)
  : DBus::ObjectAdaptor(connection, HID_SERVER_PATH)
{

    // Initialize class variables if needed
    noiseTimeout = 0; // Value of 0 means: no noise timeout

    // Initialize button timer
    btnTimer = new ButtonTimer(25,6000); // Short press should take at leas 25 ms, and a Long press takes 6 seconds
    onShortPressConnection = btnTimer->onShortPress.connect(boost::bind(&HidServer::onShortPress, this, _1));
    onLongPressConnection = btnTimer->onLongPress.connect(boost::bind(&HidServer::onLongPress, this, _1));
    onValidatePressConnection = btnTimer->onValidatePress.connect(boost::bind(&HidServer::onValidatePress, this, _1));

    // **** Initialize the GPIO Interrupt pin
    
    clog << kLogInfo << PACKAGE_STRING << " starting..." << endl;
    
    try
    {
        initHardware();
    }
    catch(std::exception x)
    {
        clog << kLogCrit << "Fatal error during hardware initialization: " << x.what() << endl << "Quitting now... " << endl;
        niam(0);
    }

    clog << kLogInfo << "Initialization complete, listening to HID requests." << endl; 
}

HidServer::~HidServer()
{
    ClearColor();

    onInterruptErrorConnection.disconnect();
    onInterruptConnection.disconnect();
    
    delete intpin; intpin = NULL;
    delete mcp; mcp = NULL;
    delete btnTimer; btnTimer = NULL;

    clog << kLogInfo << "Stopping normally" << endl;
}

void HidServer::initHardware(void)
{
    HWConfig hwConfig;
    unsigned short value;

    clog << kLogInfo << "Opening GPIO pin " << GPIO_INT_PIN << " for interrupt listening" << endl;
    // open pin
    intpin = new GpioPin(   GPIO_INT_PIN,   // Pin number
                            kDirectionIn,   // Data direction
                            kEdgeFalling);  // Interrupt edge

    // **** Initialize the MCP23017
    
    // Hardware config
    hwConfig.DISSLEW = false;   // Leave slew rate control enabled
    hwConfig.INT_MIRROR = true; // Interconnect I/O pins
    hwConfig.INT_ODR = false;   // Interrupt is not an open drain
    hwConfig.INT_POL = false;    // Interrupt is Active-Low 

    clog << kLogInfo << "Opening MCP23017 IO expander on I2C address " << MCP23017_ADR << endl;

    try
    {
        // Initialize chip system
        mcp = new Mcp23017( MCP23017_ADR,   // adr
                            0x0fff,         // iodir
                            0x0fff,         // ipol
                            0x0fff,         // pullup
                            hwConfig,       // Hardware Config
                            true);          // Swap A/B

        try
        {
            clog << kLogInfo << "Enabling interrupts on IO Expander" << endl;
            // Configure interrupt
            mcp->IntConfig( 0x0000,  // defval
                            0x0000,  // intcon
                            0x07ff); // int enable


            clog << kLogInfo << "Starting interrupt listener on GPIO pin " << GPIO_INT_PIN << endl;

            // Initialize interrupts
           
            onInterruptErrorConnection.disconnect();
            onInterruptConnection.disconnect();
            
            onInterruptErrorConnection = intpin->onThreadError.connect(boost::bind(&HidServer::onInterruptError, this, _1, _2));
            onInterruptConnection = intpin->onInterrupt.connect(boost::bind(&HidServer::onInterrupt, this, _1, _2, _3));

            // Start interrupt event on interrupt pin
            intpin->InterruptStart();

            clog << kLogInfo << "Performing initial read of IO Expander values to clear any pending interrupts and initialize Jack state" << endl;
            // Read initial value to clear any current interrupts
            value = mcp->getValue();
            
            // Pass the initial mini-jack state to any listeners
            if(value && SENSE_MINIJACK)
                JackIn();
            else
                JackOut();

            // Prepare the R/G/B pins for possible PWM usage
            mcp->setPwmState(PIN_RED,TRUE);
            mcp->setPwmState(PIN_GREEN,TRUE);
            mcp->setPwmState(PIN_BLUE,TRUE);

        }
        catch(OperationFailedException x)
        {
            delete mcp; mcp = NULL;
            throw x;
        }
            
    }
    catch(OperationFailedException x)
    {
        delete intpin; intpin = NULL;
        throw x;
    }
}

void HidServer::SetColor(const uint8_t& r, const uint8_t& g, const uint8_t& b)
{
    clog << kLogDebug << "Got color request: RGB("<< (int)r << "," << (int)g << "," << (int)b << ") - ";
    // Check if we can do solid colors (no need for PWM)
    if( (r == 0 || r==255) &&
        (g == 0 || g==255) &&
        (b == 0 || b==255))
    {
        // if so, then 0 is off and non-zero is on
        clog << "Using ON/OFF to produce color" << endl;
        // but first disable pwm
        mcp->PwmStop();   // waits for completion
        mcp->setPin(PIN_RED, (bool)r); 
        mcp->setPin(PIN_GREEN, (bool)g); 
        mcp->setPin(PIN_BLUE, (bool)b);
    
    }
    else
    {
        clog << "Using PWM to produce color" << endl;
        // we need PWM, so enable it.
        mcp->PwmStart();
      
        mcp->setPwmLedValue(PIN_RED, r); 
        mcp->setPwmLedValue(PIN_GREEN, g); 
        mcp->setPwmLedValue(PIN_BLUE, b);
    }

}

void HidServer::PulseColor(const uint8_t& r, const uint8_t& g, const uint8_t& b, const double& interval)
{
	pulseR = r;
	pulseG = g;
	pulseB = b;

	pulseDirUp = true;
	pulseMin = 0;
	pulseI = pulseMin;
	
    clog << kLogDebug << "Got pulse color request: RGB("<< (int)r << "," << (int)g << "," << (int)b << ")" << endl;
    
    pulseTime = (int32_t)((interval * 1000000) / (2 * PULSE_STEPS));
	// Start pulse thread and PWM system
	ThreadStart();
    mcp->PwmStart();

}

// contains the pulse loop
void HidServer::ThreadLoop()
{
	uint8_t myR, myG, myB;
	int32_t factor = pulseI*100;

	
	myR = (pulseR * factor)/(100 * PULSE_STEPS);
	myG = (pulseG * factor)/(100 * PULSE_STEPS);
	myB = (pulseB * factor)/(100 * PULSE_STEPS);
	
	mcp->setPwmLedValue(PIN_RED, myR); 
	mcp->setPwmLedValue(PIN_GREEN, myG); 
	mcp->setPwmLedValue(PIN_BLUE, myB);	
	
	
	if(pulseDirUp)
	{
		pulseI++;
		if(pulseI>= PULSE_STEPS)
			pulseDirUp = false;
	}
	else
	{
		pulseI--;
		if(pulseI <= pulseMin)
			pulseDirUp = true;
	}
	
	usleep(pulseTime);
		
}

void HidServer::ClearColor()
{
    clog << kLogDebug << "Got request to clear colors" << endl;

	// Stop pulse thread 
	ThreadStop();

    // Disable PWM and turn all colors off
    mcp->PwmStop();   // waits for completion
    mcp->setPin(PIN_RED, false); 
    mcp->setPin(PIN_GREEN, false); 
    mcp->setPin(PIN_BLUE, false);
}

bool HidServer::JackState()
{
    uint16_t value;
    
    try
    {
        value = mcp->getValue(); 
    }
    catch(OperationFailedException x)
    {
        // Log the exception
        clog << kLogError << "Error while retrieving Jack State: " << x.what() << endl; 
        
        // Return false on error
        return false;
    }
    // Pass the current mini-jack state to any listeners
    if(value && SENSE_MINIJACK)
        return true;
    else
        return false;
}

bool HidServer::onValidatePress(uint16_t keycode)
{
    // Check if the key is still pressed before sending out a long press

    if(mcp->getValue() & keycode)
        return true;
    else
        return false;
}

void HidServer::onShortPress(uint16_t keycode)
{
    clog << kLogDebug << "SHORT PRESS : " << getBtnName(keycode) << endl;
    ButtonPress(getBtnName(keycode));
}

void HidServer::onLongPress(uint16_t keycode)
{
    clog << kLogDebug << "LONG PRESS  : " << getBtnName(keycode) << endl;
    ButtonLongPress(getBtnName(keycode));
}

void HidServer::keyUp(uint16_t keycode)
{
    if(keycode == SENSE_MINIJACK)
    {
        JackOut();
    }
    else
    {
        ButtonUp(getBtnName(keycode));
        btnTimer->RegisterRelease(keycode);
    }
}

void HidServer::keyDown(uint16_t keycode)
{
    if(keycode == SENSE_MINIJACK)
    {
        JackIn();
    }
    else
    {
        ButtonDown(getBtnName(keycode));
        btnTimer->RegisterPress(keycode);
    }
}

// Convert binary keycode into a string for ease of use in e.g. python or other dbus bindings
std::string HidServer::getBtnName(uint16_t keycode)
{
    switch(keycode)
    {
        case BTN_ESCAPE :
            return "BTN_ESCAPE";
        case BTN_FAVORITES :
            return "BTN_FAVORITES";
        case BTN_RECENT :
            return "BTN_RECENT";

        case BTN_REWIND :
            return "BTN_REWIND";
        case BTN_PLAYPAUSE :
            return "BTN_PLAYPAUSE";
        case BTN_FASTFORWARD :
            return "BTN_FASTFORWARD";

        case BTN_CANCEL :
            return "BTN_CANCEL";
        case BTN_OK :
            return "BTN_OK";
        case BTN_UP :
            return "BTN_UP";
        case BTN_DOWN :
            return "BTN_DOWN";
        case SENSE_MINIJACK :
            return "SENSE_MINIJACK";
        default:
            return "BTN_NONE";
    }

}

void HidServer::onInterrupt(GpioPin * sender, GpioEdge edge, bool pinval)
{
    uint16_t intf,intcap, keycode;
    uint8_t i, bitcount;
    
    if(edge != kEdgeFalling)   // Only continue on falling edge
        return;


    intf = mcp->getIntF();
    intcap = mcp->getIntCap();

    // Check if we are in a noise timeout, and cancel if so.
    if(noiseTimeout !=0 && noiseTimeout > now_ms())
        return;
    else
        noiseTimeout = 0;


    // Check if only one bit is set in the interrupt cap: This indicated normal operation.
    // If more than one bit is set, we should assume noise
    bitcount = 0;
    for(i=0;i<16;i++)
    {
        if(intf & (1 << i)) // Check if this key is set
            bitcount++;
    }

    //printf("Interrupt!\n  INTF   : 0x%04x\n  INTCAP : 0x%04x\n  KEYCODE: 0x%04x\n\n", intf, intcap, keycode);
    
    if(bitcount > 0 && bitcount <= 3 ) // anything above 3 key interrupts at the same time is considered noise
    {
        
        for(i=0; i < 16; i++)
        {
            if(intf & (1 << i)) // check if this keycode is set
            {
                keycode = (1 << i);
                if(intcap & keycode)
                {
                    keyDown(keycode);
                    if(keycode == SENSE_MINIJACK)
                        clog << kLogDebug << "Jacked IN " << endl;    
                    else
                        clog << kLogDebug << "Button down : " << getBtnName(keycode) << endl;
                }
                else
                {
                    keyUp(keycode);
                    if(keycode == SENSE_MINIJACK)
                        clog << kLogDebug << "Jacked OUT " << endl;    
                    else
                        clog << kLogDebug << "Button up   : " << getBtnName(keycode) << endl;
                }
            }
        }
    }
    else
    {
        clog << kLogDebug << "!!! Input Noise !!! (Timeout: " << NOISE_TIMEOUT_MS << "ms)" << endl;
        noiseTimeout = now_ms() + NOISE_TIMEOUT_MS;
    }


}

void HidServer::onInterruptError(Thread* sender, ThreadException x)
{
    // When this function is called, the interrupt thread will have stopped
    clog << kLogErr << "Error in interrupt listener: " << x.what() << endl;
    clog << kLogErr << "Attempting to restart interrupt listener... " << endl;
    try
    {
        delete mcp; mcp = NULL;
        delete intpin; intpin = NULL;
        initHardware(); // Attempt to re-init the chip system. Quit on failure
        clog << kLogInfo << "Succesfully restarted interrupt listener" << endl;
    }
    catch(std::exception x2)
    {
        clog << kLogCrit << "Fatal error while attempting to restart interrupt listener : " << x2.what() << endl << "Quitting now... " << endl;
        niam(0);
    }
}

// get current time in milliseconds
int64_t now_ms(void)
{
    struct timespec now;
    clock_gettime(CLOCK_MONOTIC, &now);

    return ((int64_t)now.tv_sec)*1000LL + (now.tv_nsec/1000000);
}


DBus::BusDispatcher dispatcher;

void niam(int sig)
{
  dispatcher.leave();
}

int main()
{
  signal(SIGTERM, niam);
  signal(SIGINT, niam);

  DBus::default_dispatcher = &dispatcher;

  // Initialize clog to be redirected to syslog key "mediacore.hid.server"
  Log::Init("mediacore.hid");

  DBus::Connection conn = DBus::Connection::SystemBus();
  conn.request_name(HID_SERVER_NAME);

  HidServer server(conn);


  dispatcher.enter();

  return 0;
}